/*!
    \page wolfssl_API wolfSSL API Reference
    - \ref CertManager
    - \ref Memory
    - \ref openSSL

    - \ref CertsKeys
    - \ref IO
    - \ref Setup
    - \ref Debug
    - \ref TLS
*/
/*!
    \page wolfcrypt_API wolfCrypt API Reference
    <ul>
        <li>\ref ASN</li>
        <li>\ref Base_Encoding</li>
        <li>\ref Compression</li>
        <li>\ref Error</li>
        <li>\ref Keys</li>
        <li>\ref Logging</li>
        <li>\ref Math</li>
        <li>\ref Random</li>
        <li>\ref Signature</li>
        <li>\ref wolfCrypt</li>
    </ul>
    <ul>
        <li>\ref DES</li>
        <li>\ref AES</li>
        <li>\ref ARC4</li>
        <li>\ref BLAKE2</li>
        <li>\ref Camellia</li>
        <li>\ref ChaCha</li>
        <li>\ref ChaCha20Poly1305</li>
        <li>\ref Curve25519</li>
        <li>\ref DSA</li>
        <li>\ref Diffie-Hellman</li>
        <li>\ref ECC</li>
        <li>\ref ED25519</li>
        <li>\ref HC128</li>
        <li>\ref HMAC</li>
        <li>\ref IDEA</li>
        <li>\ref MD2</li>
        <li>\ref MD4</li>
        <li>\ref MD5</li>
        <li>\ref Password</li>
        <li>\ref PKCS7</li>
        <li>\ref PKCS11</li>
        <li>\ref Poly1305</li>
        <li>\ref Rabbit</li>
        <li>\ref RIPEMD</li>
        <li>\ref RSA</li>
        <li>\ref SHA</li>
        <li>\ref SRP</li>
    </ul>
*/
